"""Post-processing module."""

__author__ = "Lukas Leufen, Felix Kleinert"
__date__ = '2019-12-11'

import inspect
import logging
import os
from typing import Dict, Tuple, Union, List

import keras
import numpy as np
import pandas as pd
import xarray as xr

from mlair.data_handler import BootStraps, KerasIterator
from mlair.helpers.datastore import NameNotFoundInDataStore
from mlair.helpers import TimeTracking, statistics, extract_value
from mlair.model_modules.linear_model import OrdinaryLeastSquaredModel
from mlair.model_modules.model_class import AbstractModelClass
from mlair.plotting.postprocessing_plotting import PlotMonthlySummary, PlotStationMap, PlotClimatologicalSkillScore, \
    PlotCompetitiveSkillScore, PlotTimeSeries, PlotBootstrapSkillScore, PlotAvailability, PlotConditionalQuantiles
from mlair.run_modules.run_environment import RunEnvironment


class PostProcessing(RunEnvironment):
    """
    Perform post-processing for performance evaluation.

    Schedule of post-processing:
        #. train a ordinary least squared model (ols) for reference
        #. create forecasts for nn, ols, and persistence
        #. evaluate feature importance with bootstrapped predictions
        #. calculate skill scores
        #. create plots

    Required objects [scope] from data store:
        * `best_model` [.] or locally saved model plus `model_name` [model] and `model` [model]
        * `generator` [train, val, test, train_val]
        * `forecast_path` [.]
        * `plot_path` [postprocessing]
        * `model_path` [.]
        * `target_var` [.]
        * `sampling` [.]
        * `output_shape` [model]
        * `evaluate_bootstraps` [postprocessing] and if enabled:

            * `create_new_bootstraps` [postprocessing]
            * `bootstrap_path` [postprocessing]
            * `number_of_bootstraps` [postprocessing]

    Optional objects
        * `batch_size` [model]

    Creates
        * forecasts in `forecast_path` if enabled
        * bootstraps in `bootstrap_path` if enabled
        * plots in `plot_path`

    """

    def __init__(self):
        """Initialise and run post-processing."""
        super().__init__()
        self.model: keras.Model = self._load_model()
        self.ols_model = None
        self.batch_size: int = self.data_store.get_default("batch_size", "model", 64)
        self.test_data = self.data_store.get("data_collection", "test")
        batch_path = self.data_store.get("batch_path", scope="test")
        self.test_data_distributed = KerasIterator(self.test_data, self.batch_size, model=self.model, name="test", batch_path=batch_path)
        self.train_data = self.data_store.get("data_collection", "train")
        self.val_data = self.data_store.get("data_collection", "val")
        self.train_val_data = self.data_store.get("data_collection", "train_val")
        self.plot_path: str = self.data_store.get("plot_path")
        self.target_var = self.data_store.get("target_var")
        self._sampling = self.data_store.get("sampling")
        self.window_lead_time = extract_value(self.data_store.get("output_shape", "model"))
        self.skill_scores = None
        self.bootstrap_skill_scores = None
        self._run()

    def _run(self):
        # ols model
        self.train_ols_model()

        # forecasts
        self.make_prediction()

        # skill scores on test data
        self.calculate_test_score()

        # bootstraps
        if self.data_store.get("evaluate_bootstraps", "postprocessing"):
            with TimeTracking(name="calculate bootstraps"):
                create_new_bootstraps = self.data_store.get("create_new_bootstraps", "postprocessing")
                self.bootstrap_postprocessing(create_new_bootstraps)

        # skill scores
        with TimeTracking(name="calculate skill scores"):
            self.skill_scores = self.calculate_skill_scores()

        # plotting
        self.plot()

    def bootstrap_postprocessing(self, create_new_bootstraps: bool, _iter: int = 0) -> None:
        """
        Calculate skill scores of bootstrapped data.

        Create bootstrapped data if create_new_bootstraps is true or a failure occurred during skill score calculation
        (this will happen by default, if no bootstrapped data is available locally). Set class attribute
        bootstrap_skill_scores. This method is implemented in a recursive fashion, but is only allowed to call itself
        once.

        :param create_new_bootstraps: calculate all bootstrap predictions and overwrite already available predictions
        :param _iter: internal counter to reduce unnecessary recursive calls (maximum number is 2, otherwise something
            went wrong).
        """
        try:
            if create_new_bootstraps:
                self.create_bootstrap_forecast()
            self.bootstrap_skill_scores = self.calculate_bootstrap_skill_scores()
        except FileNotFoundError:
            if _iter != 0:
                raise RuntimeError("bootstrap_postprocessing is called for the 2nd time. This means, that calling"
                                   "manually the reason for the failure.")
            logging.info("Couldn't load all files, restart bootstrap postprocessing with create_new_bootstraps=True.")
            self.bootstrap_postprocessing(True, _iter=1)

    def create_bootstrap_forecast(self) -> None:
        """
        Create bootstrapped predictions for all stations and variables.

        These forecasts are saved in bootstrap_path with the names `bootstraps_{var}_{station}.nc` and
        `bootstraps_labels_{station}.nc`.
        """
        # forecast
        with TimeTracking(name=inspect.stack()[0].function):
            # extract all requirements from data store
            bootstrap_path = self.data_store.get("bootstrap_path")
            forecast_path = self.data_store.get("forecast_path")
            number_of_bootstraps = self.data_store.get("number_of_bootstraps", "postprocessing")
            dims = ["index", "ahead", "type"]
            for station in self.test_data:
                logging.info(str(station))
                X, Y = None, None
                bootstraps = BootStraps(station, number_of_bootstraps)
                for boot in bootstraps:
                    X, Y, (index, dimension) = boot
                    # make bootstrap predictions
                    bootstrap_predictions = self.model.predict(X)
                    if isinstance(bootstrap_predictions, list):  # if model is branched model
                        bootstrap_predictions = bootstrap_predictions[-1]
                    # save bootstrap predictions separately for each station and variable combination
                    bootstrap_predictions = np.expand_dims(bootstrap_predictions, axis=-1)
                    shape = bootstrap_predictions.shape
                    coords = (range(shape[0]), range(1, shape[1] + 1))
                    var = f"{index}_{dimension}"
                    tmp = xr.DataArray(bootstrap_predictions, coords=(*coords, [var]), dims=dims)
                    file_name = os.path.join(forecast_path, f"bootstraps_{station}_{var}.nc")
                    tmp.to_netcdf(file_name)
                else:
                    # store also true labels for each station
                    labels = np.expand_dims(Y, axis=-1)
                    file_name = os.path.join(forecast_path, f"bootstraps_{station}_labels.nc")
                    labels = xr.DataArray(labels, coords=(*coords, ["obs"]), dims=dims)
                    labels.to_netcdf(file_name)

    def calculate_bootstrap_skill_scores(self) -> Dict[str, xr.DataArray]:
        """
        Calculate skill score of bootstrapped variables.

        Use already created bootstrap predictions and the original predictions (the not-bootstrapped ones) and calculate
        skill scores for the bootstraps. The result is saved as a xarray DataArray in a dictionary structure separated
        for each station (keys of dictionary).

        :return: The result dictionary with station-wise skill scores
        """
        with TimeTracking(name=inspect.stack()[0].function):
            # extract all requirements from data store
            bootstrap_path = self.data_store.get("bootstrap_path")
            forecast_path = self.data_store.get("forecast_path")
            number_of_bootstraps = self.data_store.get("number_of_bootstraps", "postprocessing")
            forecast_file = f"forecasts_norm_%s_test.nc"
            bootstraps = BootStraps(self.test_data[0], number_of_bootstraps).bootstraps()
            skill_scores = statistics.SkillScores(None)
            score = {}
            for station in self.test_data:
                logging.info(station)

                # get station labels
                file_name = os.path.join(forecast_path, f"bootstraps_{str(station)}_labels.nc")
                labels = xr.open_dataarray(file_name)
                shape = labels.shape

                # get original forecasts
                orig = self.get_orig_prediction(forecast_path, forecast_file % str(station), number_of_bootstraps)
                orig = orig.reshape(shape)
                coords = (range(shape[0]), range(1, shape[1] + 1), ["orig"])
                orig = xr.DataArray(orig, coords=coords, dims=["index", "ahead", "type"])

                # calculate skill scores for each variable
                skill = pd.DataFrame(columns=range(1, self.window_lead_time + 1))
                for boot_set in bootstraps:
                    boot_var = f"{boot_set[0]}_{boot_set[1]}"
                    file_name = os.path.join(forecast_path, f"bootstraps_{station}_{boot_var}.nc")
                    boot_data = xr.open_dataarray(file_name)
                    boot_data = boot_data.combine_first(labels).combine_first(orig)
                    boot_scores = []
                    for ahead in range(1, self.window_lead_time + 1):
                        data = boot_data.sel(ahead=ahead)
                        boot_scores.append(
                            skill_scores.general_skill_score(data, forecast_name=boot_var, reference_name="orig"))
                    skill.loc[boot_var] = np.array(boot_scores)

                # collect all results in single dictionary
                score[str(station)] = xr.DataArray(skill, dims=["boot_var", "ahead"])
            return score

    @staticmethod
    def get_orig_prediction(path, file_name, number_of_bootstraps, prediction_name="CNN"):
        file = os.path.join(path, file_name)
        prediction = xr.open_dataarray(file).sel(type=prediction_name).squeeze()
        vals = np.tile(prediction.data, (number_of_bootstraps, 1))
        return vals[~np.isnan(vals).any(axis=1), :]

    def _load_model(self) -> keras.models:
        """
        Load NN model either from data store or from local path.

        :return: the model
        """
        try:
            model = self.data_store.get("best_model")
        except NameNotFoundInDataStore:
            logging.info("No model was saved in data store. Try to load model from experiment path.")
            model_name = self.data_store.get("model_name", "model")
            model_class: AbstractModelClass = self.data_store.get("model", "model")
            model = keras.models.load_model(model_name, custom_objects=model_class.custom_objects)
        return model

    def plot(self):
        """
        Create all plots.

        Plots are defined in experiment set up by `plot_list`. As default, all (following) plots are enabled:

        * :py:class:`PlotBootstrapSkillScore <src.plotting.postprocessing_plotting.PlotBootstrapSkillScore>`
        * :py:class:`PlotConditionalQuantiles <src.plotting.postprocessing_plotting.PlotConditionalQuantiles>`
        * :py:class:`PlotStationMap <src.plotting.postprocessing_plotting.PlotStationMap>`
        * :py:class:`PlotMonthlySummary <src.plotting.postprocessing_plotting.PlotMonthlySummary>`
        * :py:class:`PlotClimatologicalSkillScore <src.plotting.postprocessing_plotting.PlotClimatologicalSkillScore>`
        * :py:class:`PlotCompetitiveSkillScore <src.plotting.postprocessing_plotting.PlotCompetitiveSkillScore>`
        * :py:class:`PlotTimeSeries <src.plotting.postprocessing_plotting.PlotTimeSeries>`
        * :py:class:`PlotAvailability <src.plotting.postprocessing_plotting.PlotAvailability>`

        .. note:: Bootstrap plots are only created if bootstraps are evaluated.

        """
        logging.debug("Run plotting routines...")
        path = self.data_store.get("forecast_path")

        plot_list = self.data_store.get("plot_list", "postprocessing")
        time_dimension = self.data_store.get("time_dim")

        if self.bootstrap_skill_scores is not None and "PlotBootstrapSkillScore" in plot_list:
            PlotBootstrapSkillScore(self.bootstrap_skill_scores, plot_folder=self.plot_path, model_setup="CNN")

        if "PlotConditionalQuantiles" in plot_list:
            PlotConditionalQuantiles(self.test_data.keys(), data_pred_path=path, plot_folder=self.plot_path)
        if "PlotStationMap" in plot_list:
            if self.data_store.get("hostname")[:2] in self.data_store.get("hpc_hosts") or self.data_store.get(
                    "hostname")[:6] in self.data_store.get("hpc_hosts"):
                logging.warning(
                    f"Skip 'PlotStationMap` because running on a hpc node: {self.data_store.get('hostname')}")
            else:
                PlotStationMap(generators={'b': self.test_data}, plot_folder=self.plot_path)
        if "PlotMonthlySummary" in plot_list:
            PlotMonthlySummary(self.test_data.keys(), path, r"forecasts_%s_test.nc", self.target_var,
                               plot_folder=self.plot_path)
        if "PlotClimatologicalSkillScore" in plot_list:
            PlotClimatologicalSkillScore(self.skill_scores[1], plot_folder=self.plot_path, model_setup="CNN")
            PlotClimatologicalSkillScore(self.skill_scores[1], plot_folder=self.plot_path, score_only=False,
                                         extra_name_tag="all_terms_", model_setup="CNN")
        if "PlotCompetitiveSkillScore" in plot_list:
            PlotCompetitiveSkillScore(self.skill_scores[0], plot_folder=self.plot_path, model_setup="CNN")
        if "PlotTimeSeries" in plot_list:
            PlotTimeSeries(self.test_data.keys(), path, r"forecasts_%s_test.nc", plot_folder=self.plot_path,
                           sampling=self._sampling)
        if "PlotAvailability" in plot_list:
            avail_data = {"train": self.train_data, "val": self.val_data, "test": self.test_data}
            PlotAvailability(avail_data, plot_folder=self.plot_path, time_dimension=time_dimension)

    def calculate_test_score(self):
        """Evaluate test score of model and save locally."""
        test_score = self.model.evaluate_generator(generator=self.test_data_distributed,
                                                   use_multiprocessing=True, verbose=0, steps=1)
        path = self.data_store.get("model_path")
        with open(os.path.join(path, "test_scores.txt"), "a") as f:
            for index, item in enumerate(test_score):
                logging.info(f"{self.model.metrics_names[index]}, {item}")
                f.write(f"{self.model.metrics_names[index]}, {item}\n")

    def train_ols_model(self):
        """Train ordinary least squared model on train data."""
        self.ols_model = OrdinaryLeastSquaredModel(self.train_data)

    def make_prediction(self):
        """
        Create predictions for NN, OLS, and persistence and add true observation as reference.

        Predictions are filled in an array with full index range. Therefore, predictions can have missing values. All
        predictions for a single station are stored locally under `<forecast/forecast_norm>_<station>_test.nc` and can
        be found inside `forecast_path`.
        """
        logging.debug("start make_prediction")
        time_dimension = self.data_store.get("time_dim")
        for i, data in enumerate(self.test_data):
            input_data = data.get_X()
            target_data = data.get_Y(as_numpy=False)
            observation_data = data.get_observation()

            # get scaling parameters
            mean, std, transformation_method = data.get_transformation_Y()

            for normalised in [True, False]:
                # create empty arrays
                nn_prediction, persistence_prediction, ols_prediction, observation = self._create_empty_prediction_arrays(
                    target_data, count=4)

                # nn forecast
                nn_prediction = self._create_nn_forecast(input_data, nn_prediction, mean, std, transformation_method,
                                                         normalised)

                # persistence
                persistence_prediction = self._create_persistence_forecast(observation_data, persistence_prediction, mean, std,
                                                                           transformation_method, normalised)

                # ols
                ols_prediction = self._create_ols_forecast(input_data, ols_prediction, mean, std, transformation_method,
                                                           normalised)

                # observation
                observation = self._create_observation(target_data, observation, mean, std, transformation_method, normalised)

                # merge all predictions
                full_index = self.create_fullindex(observation_data.indexes[time_dimension], self._get_frequency())
                all_predictions = self.create_forecast_arrays(full_index, list(target_data.indexes['window']),
                                                              time_dimension,
                                                              CNN=nn_prediction,
                                                              persi=persistence_prediction,
                                                              obs=observation,
                                                              OLS=ols_prediction)

                # save all forecasts locally
                path = self.data_store.get("forecast_path")
                prefix = "forecasts_norm" if normalised else "forecasts"
                file = os.path.join(path, f"{prefix}_{str(data)}_test.nc")
                all_predictions.to_netcdf(file)

    def _get_frequency(self) -> str:
        """Get frequency abbreviation."""
        getter = {"daily": "1D", "hourly": "1H"}
        return getter.get(self._sampling, None)

    @staticmethod
    def _create_observation(data, _, mean: xr.DataArray, std: xr.DataArray, transformation_method: str,
                            normalised: bool) -> xr.DataArray:
        """
        Create observation as ground truth from given data.

        Inverse transformation is applied to the ground truth to get the output in the original space.

        :param data: observation
        :param mean: mean of target value transformation
        :param std: standard deviation of target value transformation
        :param transformation_method: target values transformation method
        :param normalised: transform ground truth in original space if false, or use normalised predictions if true

        :return: filled data array with observation
        """
        if not normalised:
            data = statistics.apply_inverse_transformation(data, mean, std, transformation_method)
        return data

    def _create_ols_forecast(self, input_data: xr.DataArray, ols_prediction: xr.DataArray, mean: xr.DataArray,
                             std: xr.DataArray, transformation_method: str, normalised: bool) -> xr.DataArray:
        """
        Create ordinary least square model forecast with given input data.

        Inverse transformation is applied to the forecast to get the output in the original space.

        :param data: transposed history from DataPrep
        :param ols_prediction: empty array in right shape to fill with data
        :param mean: mean of target value transformation
        :param std: standard deviation of target value transformation
        :param transformation_method: target values transformation method
        :param normalised: transform prediction in original space if false, or use normalised predictions if true

        :return: filled data array with ols predictions
        """
        tmp_ols = self.ols_model.predict(input_data)
        if not normalised:
            tmp_ols = statistics.apply_inverse_transformation(tmp_ols, mean, std, transformation_method)
        target_shape = ols_prediction.values.shape
        ols_prediction.values = np.swapaxes(tmp_ols, 2, 0) if target_shape != tmp_ols.shape else tmp_ols
        return ols_prediction

    def _create_persistence_forecast(self, data, persistence_prediction: xr.DataArray, mean: xr.DataArray,
                                     std: xr.DataArray, transformation_method: str, normalised: bool) -> xr.DataArray:
        """
        Create persistence forecast with given data.

        Persistence is deviated from the value at t=0 and applied to all following time steps (t+1, ..., t+window).
        Inverse transformation is applied to the forecast to get the output in the original space.

        :param data: observation
        :param persistence_prediction: empty array in right shape to fill with data
        :param mean: mean of target value transformation
        :param std: standard deviation of target value transformation
        :param transformation_method: target values transformation method
        :param normalised: transform prediction in original space if false, or use normalised predictions if true

        :return: filled data array with persistence predictions
        """
        tmp_persi = data.copy()
        if not normalised:
            tmp_persi = statistics.apply_inverse_transformation(tmp_persi, mean, std, transformation_method)
        persistence_prediction.values = np.tile(tmp_persi, (self.window_lead_time, 1)).T
        return persistence_prediction

    def _create_nn_forecast(self, input_data: xr.DataArray, nn_prediction: xr.DataArray, mean: xr.DataArray,
                            std: xr.DataArray, transformation_method: str, normalised: bool) -> xr.DataArray:
        """
        Create NN forecast for given input data.

        Inverse transformation is applied to the forecast to get the output in the original space. Furthermore, only the
        output of the main branch is returned (not all minor branches, if the network has multiple output branches). The
        main branch is defined to be the last entry of all outputs.

        :param input_data: transposed history from DataPrep
        :param nn_prediction: empty array in right shape to fill with data
        :param mean: mean of target value transformation
        :param std: standard deviation of target value transformation
        :param transformation_method: target values transformation method
        :param normalised: transform prediction in original space if false, or use normalised predictions if true

        :return: filled data array with nn predictions
        """
        tmp_nn = self.model.predict(input_data)
        if not normalised:
            tmp_nn = statistics.apply_inverse_transformation(tmp_nn, mean, std, transformation_method)
        if isinstance(tmp_nn, list):
            nn_prediction.values = tmp_nn[-1]
        elif tmp_nn.ndim == 3:
            nn_prediction.values = tmp_nn[-1, ...]
        elif tmp_nn.ndim == 2:
            nn_prediction.values = tmp_nn
        else:
            raise NotImplementedError(f"Number of dimension of model output must be 2 or 3, but not {tmp_nn.dims}.")
        return nn_prediction

    @staticmethod
    def _create_empty_prediction_arrays(target_data, count=1):
        """
        Create array to collect all predictions. Expand target data by a station dimension. """
        return [target_data.copy() for _ in range(count)]

    @staticmethod
    def create_fullindex(df: Union[xr.DataArray, pd.DataFrame, pd.DatetimeIndex], freq: str) -> pd.DataFrame:
        """
        Create full index from first and last date inside df and resample with given frequency.

        :param df: use time range of this data set
        :param freq: frequency of full index

        :return: empty data frame with full index.
        """
        if isinstance(df, pd.DataFrame):
            earliest = df.index[0]
            latest = df.index[-1]
        elif isinstance(df, xr.DataArray):
            earliest = df.index[0].values
            latest = df.index[-1].values
        elif isinstance(df, pd.DatetimeIndex):
            earliest = df[0]
            latest = df[-1]
        else:
            raise AttributeError(f"unknown array type. Only pandas dataframes, xarray dataarrays and pandas datetimes "
                                 f"are supported. Given type is {type(df)}.")
        index = pd.DataFrame(index=pd.date_range(earliest, latest, freq=freq))
        return index

    @staticmethod
    def create_forecast_arrays(index: pd.DataFrame, ahead_names: List[Union[str, int]], time_dimension, **kwargs):
        """
        Combine different forecast types into single xarray.

        :param index: index for forecasts (e.g. time)
        :param ahead_names: names of ahead values (e.g. hours or days)
        :param kwargs: as xarrays; data of forecasts

        :return: xarray of dimension 3: index, ahead_names, # predictions

        """
        keys = list(kwargs.keys())
        res = xr.DataArray(np.full((len(index.index), len(ahead_names), len(keys)), np.nan),
                           coords=[index.index, ahead_names, keys], dims=['index', 'ahead', 'type'])
        for k, v in kwargs.items():
            match_index = np.stack(set(res.index.values) & set(v.indexes[time_dimension].values))
            res.loc[match_index, :, k] = v.loc[match_index]
        return res

    def _get_external_data(self, station: str) -> Union[xr.DataArray, None]:
        """
        Get external data for given station.

        External data is defined as data that is not part of the observed period. From an evaluation perspective, this
        refers to data, that is no test data, and therefore to train and val data.

        :param station: name of station to load external data.
        """
        try:
            data = self.train_val_data[station]
            # target_data = data.get_Y(as_numpy=False)
            observation = data.get_observation()
            mean, std, transformation_method = data.get_transformation_Y()
            # external_data = self._create_observation(target_data, None, mean, std, transformation_method, normalised=False)
            # external_data = external_data.squeeze("Stations").sel(window=1).drop(["window", "Stations", "variables"])
            external_data = self._create_observation(observation, None, mean, std, transformation_method, normalised=False)
            return external_data.rename({external_data.dims[0]: 'index'})
        except IndexError:
            return None

    def calculate_skill_scores(self) -> Tuple[Dict, Dict]:
        """
        Calculate skill scores of CNN forecast.

        The competitive skill score compares the CNN prediction with persistence and ordinary least squares forecasts.
        Whereas, the climatological skill scores evaluates the CNN prediction in terms of meaningfulness in comparison
        to different climatological references.

        :return: competitive and climatological skill scores
        """
        path = self.data_store.get("forecast_path")
        skill_score_competitive = {}
        skill_score_climatological = {}
        for station in self.test_data:
            file = os.path.join(path, f"forecasts_{str(station)}_test.nc")
            data = xr.open_dataarray(file)
            skill_score = statistics.SkillScores(data)
            external_data = self._get_external_data(station)
            skill_score_competitive[station] = skill_score.skill_scores(self.window_lead_time)
            skill_score_climatological[station] = skill_score.climatological_skill_scores(external_data,
                                                                                          self.window_lead_time)
        return skill_score_competitive, skill_score_climatological
